import numpy as np
from geopandas import GeoDataFrame
from pandas import Series
from pylusat.base import GeoDataFrameManager
from pylusat.base import RasterManager
from pylusat.utils import rasterize_geometry, cntrd_array, inv_affine


class _ArrayDistance:

    SUPPORT_DIST = ["Point", "Line", "Raster"]

    def __init__(self, target_type, method, dtype):
        if target_type.title() not in self.SUPPORT_DIST:
            raise ValueError('Unsupported target data. Must be "Point", '
                             '"Line", or "Raster".')
        self.target_type = target_type
        self.method = method
        self.dtype = dtype

    @staticmethod
    def _kdtree(target_arr):
        from scipy.spatial import cKDTree
        return cKDTree(target_arr)

    def query(self, source_arr, target_arr):
        # query distance from source_arr to target_arr
        kdtree = self._kdtree(target_arr)
        dist_method = 1 if self.method.lower() == 'manhattan' else 2
        dist_arr = kdtree.query(source_arr, p=dist_method)[0]
        if self.dtype is not float:
            return dist_arr.astype(self.dtype)
        else:
            return dist_arr


def _validate_target_geom(gdf, geom_type):
    gdf_manager = GeoDataFrameManager(gdf)
    assert gdf_manager.geom_type_validate(geom_type), (
        "The target geometry must be {}.".format(geom_type)
    )

        
def to_point(input_gdf, point_gdf, method='euclidean', dtype=float):
    """
    Calculate distance for each geometry in the input GeoDataFrame to its
    nearest neighbor in the point GeoDataFrame.

    Parameters
    ----------
    input_gdf : GeoDataFrame
        Input GeoDataFrame. Centroids of the input geometries are used.
    point_gdf : GeoDataFrame
        The GeoDataFrame contains the point geometries to which distances are
        calculated.
    method : str, optional
        Method used to calculate distances. Either 'euclidean' or 'manhattan'.
    dtype : str or np.dtype, optional
        Use a np.dtype or Python type to cast the output distance to the
        desired type.

    Returns
    -------
    Series
        A pandas Series containing the distances of each input feature to its
        nearest point.
    """
    target_geom = "Point"
    _validate_target_geom(point_gdf, target_geom)
    pnt_dist = _ArrayDistance(target_geom, method, dtype)
    input_arr = cntrd_array(input_gdf)
    point_arr = cntrd_array(point_gdf)
    pnt_dist_arr = pnt_dist.query(input_arr, point_arr)
    return Series(pnt_dist_arr, index=input_gdf.index)


def to_line(input_gdf, line_gdf, cellsize=30, method="euclidean", dtype=float):
    """
    Calculate distances from input_gdf to line_gdf.

    Parameters
    ----------
    input_gdf : GeoDataFrame
        Input GeoDataFrame. Centroids of the input geometries are used.
    line_gdf : GeoDataFrame
        A GeoDataFrame whose geometry is of line. 
    cellsize : float
        Cell size used to rasterize the line_gdf. 
    method : str, optional
        Method used to calculate distances. Either 'euclidean' or 'manhattan'.
    dtype : str or np.dtype, optional
        Use a np.dtype or Python type to cast the output distance to the
        desired type.

    Returns
    -------
    Series
        A pandas Series of distances from each feature in input_gdf to its
        nearest neighbor in line_gdf.
    Notes
    -----
    To rapidly query distances, the line_gdf is burned into numpy array by
    using rasterize function from the rasterio package.
    """
    target_geom = "Line"
    _validate_target_geom(line_gdf, target_geom)
    line_grid, _, extent, nodata = rasterize_geometry(line_gdf, cellsize)

    max_y = extent[3]  # max_y from extent
    min_x = extent[0]  # min_x from extent

    input_arr = inv_affine(input_gdf, cellsize, max_y=max_y, min_x=min_x)
    line_arr = np.argwhere(line_grid != nodata)

    line_dist = _ArrayDistance("Line", method, dtype)
    line_dist_arr = line_dist.query(input_arr, line_arr) * cellsize

    return Series(line_dist_arr, index=input_gdf.index)


def to_cell(input_gdf, raster, value, nodata=None,
            method="euclidean", dtype=float):
    """
    Calculate distance for each geometry to its nearest-neighbor cell that has
    a specific value.

    Parameters
    ----------
    input_gdf : GeoDataFrame
        Input GeoDataFrame
    raster : str
        A path to a tif file or a connection string to a raster on PostgreSQL.
    value : int or float
        Cells in the raster with this value will be used as targets for
        distance calculation.
    nodata : int or float
        Value for no data cells.
    method : str, optional, default "euclidean"
        Method used to calculate distances. Either 'euclidean' or 'manhattan'.
    dtype : str or np.dtype, optional
        Use a np.dtype or Python type to cast the output distance to the
        desired type.

    Returns
    -------
    Series
        A pandas Series of distances from each feature in input_gdf to the
        nearest cell (has the specified value) in the raster dataset.
    """
    rast_manager = RasterManager(raster, nodata)
    rast_arr, cellsize, max_y, min_x, nodata = rast_manager.as_rebuild_info()
    rast_arr_filtered = np.argwhere(rast_arr == value)
    # if raster does not contain any specified value return null for each row
    if rast_arr_filtered.size == 0:
        return Series(np.nan, index=input_gdf.index)

    input_centroid_arr = inv_affine(input_gdf, cellsize, max_y, min_x)
    rast_dist_obj = _ArrayDistance("Raster", method, dtype)
    rast_dist = (
        rast_dist_obj.query(input_centroid_arr, rast_arr_filtered) * cellsize
    )
    return Series(rast_dist, index=input_gdf.index)



import geopandas as gpd
from pylusat.distance import to_point, to_line, to_cell
from pylusat.datasets import get_path
import pytest


@pytest.fixture
def schools_gdf():
    return gpd.read_file(get_path("schools"))


@pytest.fixture
def acs2016_gdf():
    return gpd.read_file(get_path("acs2016"))


@pytest.fixture
def highway_gdf():
    return gpd.read_file(get_path("highway"))


@pytest.fixture
def habitat_tif():
    return get_path("habitat")


def test_to_point(acs2016_gdf, schools_gdf):
    result = to_point(acs2016_gdf, schools_gdf)
    assert round(result[0], 4) == 197.2841


def test_to_line(acs2016_gdf, highway_gdf):
    result = to_line(acs2016_gdf, highway_gdf)
    assert round(result[0], 4) == 715.6116


def test_to_cell(acs2016_gdf, habitat_tif):
    result = to_cell(acs2016_gdf, habitat_tif, 6)
    assert round(result[0], 4) == 5825.4099



from rasterio import features
from affine import Affine
import numpy as np
from geopandas import GeoDataFrame
from pandas import DataFrame
from shapely.geometry import Polygon


def rasterize_geometry(gdf, cellsize, value_clm=None, value_fill=0):
    """
    Transform vector data into a 2-d array. If any (or a part of) geometry
    presents at a given cell, the cell will be assigned to a value of 1,
    otherwise 0.

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame.
    cellsize : int or float
        Cell size used to transform the vector data.
    value_clm : str, optional
        The name of the column which contains the values for each cell in the
        output array.
    value_fill : int, optional
        Fill value for all areas not covered by the input geometries.
    Returns
    -------
    arr : np.ndarray
        Output numpy array.
    trans : Affine
        An object of the Affine class.
    extent : tuple
        a tuple containing ``min_x``, ``min_y``, ``max_x``, ``max_y``
        values for the bounds of the series as a whole.
        See GeoSeries.bounds for the bounds of the geometries contained in
        the series.
    nodata : int or float
        The no data value used during the rasterization.
    """
    extent = gdf.total_bounds
    output_shape = (int(round((extent[3] - extent[1]) / cellsize)),
                    int(round((extent[2] - extent[0]) / cellsize)))
    trans = Affine(cellsize, 0, extent[0], 0, -cellsize, extent[3])
    if value_clm is None:
        arr = features.rasterize(gdf[gdf.geometry.name],
                                 out_shape=output_shape,
                                 fill=value_fill,
                                 transform=trans)
    else:
        arr = features.rasterize(
            tuple(zip(gdf[gdf.geometry.name], gdf[value_clm])),
            out_shape=output_shape, fill=value_fill, transform=trans
        )
    nodata = value_fill
    return arr, trans, extent, nodata


def cntrd_array(gdf):
    """
    Get the coordinates of the centroids in a 2D array.

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame.
    Returns
    -------
    output : np.ndarray
        An n by 2 2D array where each row contains the coordinates (x and y)
        of the centroids of each geometry in the input GeoDataFrame.
    """
    if isinstance(gdf, GeoDataFrame):
        return np.column_stack((gdf.centroid.x.values,
                                gdf.centroid.y.values))
    else:
        raise TypeError("The input data must be a GeoDataFrame.")


def inv_affine(gdf, cellsize, max_y, min_x):
    """
    Convert (x, y) coordinates of the centroids of a GeoDataFrame to
    (row, column) on a grid.

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame for the function.
    cellsize : int or float
        Cell size used to grid the vector data.
    max_y : int or float
        Upper bound on the y axis of the grid.
    min_x : int or float
        Lower bound on the x axis of the grid.

    Returns
    -------
    np.ndarray
        An n by 2 2D-array of (row, column) indices for the centroid of each
        geometry in the input GeoDataFrame.
    """
    return np.column_stack(
        (np.round((max_y - gdf.centroid.y) / cellsize).values,   # row
         np.round((min_x - gdf.centroid.x) / -cellsize).values)  # column
    )


def _calc_ahp(r_mtx):
    # random consistency index based on saaty 1990
    ri_dict = {3: 0.58, 4: 0.9, 5: 1.12, 6: 1.24,
               7: 1.32, 8: 1.41, 9: 1.45, 10: 1.49}
    n = r_mtx.shape[0]
    eig_val, eig_vec = np.linalg.eig(r_mtx)
    # principal eigenvalue, i.e., the highest valued eigenvalue
    principal_eig_val = np.max(eig_val).real
    principal_eig_vec = eig_vec[:, eig_val.argmax()].real

    ci = (principal_eig_val - n) / (n - 1)   # consistency index
    cr = ci / ri_dict[n]                     # consistency ratio

    # the normalized principal eigenvector is called the priority vector
    priority_vec = principal_eig_vec / np.sum(principal_eig_vec)
    return priority_vec, cr


def ahp(r_mtx):
    """
    Compute the priority vector of the Analytic Hierarchy Process (AHP).

    The computed priority vector can be used as weights in a decision-making
    process based on multiple criteria. The function also returns the
    Consistency Ratio (CR) to check whether decisions of pair-wise comparisons
    are consistent.

    Parameters
    ----------
    r_mtx : np.ndarray
        The reciprocal matrix.

    Returns
    -------
    priority_vec : np.ndarray
        The priority vector computed by AHP.
    cr : float
        Consistency Ratio (1990, Saaty). The rule of thumb is CR <= 0.1.
    """
    assert r_mtx.ndim == 2, "The reciprocal matrix must be 2-D."
    nrow, ncol = r_mtx.shape
    assert nrow == ncol, "The reciprocal matrix must be a square matrix."
    # validate the number of criteria is between 3 and 10
    assert 3 <= nrow <= 10, (
        "The number of rows/columns must be between 3 and 10."
    )
    # validate matrix's elements are valid members of the AHP fundamental scale
    ahp_scale = np.append(1 / (np.arange(9)+1), np.arange(1, 9)+1)
    assert (
        np.isin(r_mtx[np.tril_indices(nrow)], ahp_scale).all().item() is True
    ), ("One or more elements of the input matrix "
        "is not a valid member of "
        "the AHP fundamental scale.")
    # validate the matrix is indeed reciprocal
    assert (
        1 / r_mtx.T[np.tril_indices(nrow)] ==
        r_mtx[np.tril_indices(nrow)]
    ).all().item() is True, "Invalid reciprocal matrix."
    return _calc_ahp(r_mtx)


def random_ahp(n=3):
    """
    Generate randomized Analytic Hierarchy Process (AHP) weights.

    This function generates random weights, i.e., priority vector, using the
    AHP method (1990, Saaty). The randomly generated weights are ensured to
    be consistent by evaluating the consistency ratio (CR).

    Parameters
    ----------
    n : int, optional, default 3
        Number of AHP weights to generate, which must be between 3 and 10.

    Returns
    -------
    priority_vec : np.ndarray
        An array that contains the weights (priority vector of AHP).
    """
    if n < 3 or n > 10:
        raise ValueError("The number must be between 3 and 10.")

    # an array of scales for pair-wise comparison
    ahp_scale = np.append(1 / (np.arange(9) + 1), np.arange(1, 9) + 1)

    while True:
        # upper triangular part of AHP's reciprocal matrix
        triu_weights = np.random.choice(ahp_scale, np.sum(np.arange(n)))

        r_mtx = np.identity(n)   # the reciprocal matrix
        r_mtx[np.triu_indices(n, 1)] = triu_weights
        r_mtx[np.tril_indices(n, -1)] = 1 / r_mtx.T[np.tril_indices(n, -1)]

        priority_vec, cr = _calc_ahp(r_mtx)
        if cr < 0.1:       # general rule of thumb: CR less than 10%
            break
    return priority_vec, cr


def weighted_sum(df, col_weights):
    """
    Calculate a weighted sum over a list of existing columns.

    Parameters
    ----------
    df : DataFrame or GeoDataFrame
        Input DataFrame contains the columns for summation.
    col_weights : dict
        Dict of ``{column name: float}``, where float is the weight for the
        column specified by the column name.
    Returns
    -------
    weighted_sum : np.ndarray
        An array contains the result of the weighted sum.
    """
    wgt_df = DataFrame([*col_weights.values()], index=[*col_weights])
    return df[[*col_weights]].dot(wgt_df)


def gridify(input_gdf, width=None, height=None, num_cols=None, num_rows=None):
    """
    Create a grid based on the input_gdf by specifying width and/or height of
    the cells of the grid.

    num_cols and num_rows can be used to define the grid as well. If nothing
    specified, the cell size of the grid will be the span on x axis (width)
    divided by 30.

    Parameters
    ----------
    input_gdf : GeoDataFrame
        Input GeoDataFrame based on which the grid is created.
    width : int or float
        Cell width.
    height : int or float
        Cell height.
    num_cols : int
        Number of columns. When specified, this number will determine cell
        width.
    num_rows : int
        Number of rows. When specified, this number will determine cell height.

    Returns
    -------
    GeoDataFrame
        The output grid (polygons).
    """
    xmin, ymin, xmax, ymax = input_gdf.total_bounds

    if width and height:
        pass
    else:
        if not width and not height:
            width = height = (xmax - xmin) / 30
        else:
            if width:
                height = width
            elif height:
                width = height

    if num_cols and num_rows:
        # width and height are calculated based on num_cols and num_rows
        width = (xmax - xmin) / num_cols
        height = (ymax - ymin) / num_rows
    else:
        if num_cols:
            width = height = (xmax - xmin) / num_cols
        elif num_rows:
            height = width = (ymax - ymin) / num_rows
        else:
            pass

    cols = np.arange(xmin - width/2, xmax + width/2, width)
    rows = np.arange(ymin - height/2, ymax + height/2, height)

    polygons = [
        Polygon([
            (x, y),
            (x+width, y),
            (x+width, y+height),
            (x, y+height)
        ])
        for y in rows
        for x in cols
    ]

    return GeoDataFrame({'geometry': polygons}, crs=input_gdf.crs)



import geopandas as gpd
from pylusat.zonal import zonal_stats_raster
import pytest
from pylusat.datasets import get_path


@pytest.fixture
def acs2016_gdf():
    return gpd.read_file(get_path("acs2016"))


@pytest.fixture
def habitat_tif():
    return get_path("habitat")


def test_zonal_stats_raster(acs2016_gdf, habitat_tif):
    zonal_result = zonal_stats_raster(acs2016_gdf, habitat_tif)
    assert zonal_result.iloc[0, -1] == 1303
    assert round(zonal_result.iloc[0, -2], 4) == 32.4351
    assert zonal_result.iloc[0, -3] == 42
    assert zonal_result.iloc[0, -4] == 7



import numpy as np
import pandas as pd
from pandas import DataFrame
from geopandas import GeoDataFrame
from scipy import stats


def reclassify(input_df, input_col, reclassify_def, output_col, nodata=None):
    """
    Reclassify values in an existing column based on key-value pairs provided
    in a dictionary.

    The function can handle both categorical and interval definitions. For
    interval definition, the keys of the dictionary should be a tuple of two
    numbers corresponding to the start and end of each interval. The intervals
    are right closed.

    Parameters
    ----------
    input_df : DataFrame or GeoDataFrame
        Input DataFrame with the column need to be reclassified.
    input_col : str
        The name of the input column containing the old values.
    reclassify_def : dict
        The dictionary consists of definitions to convert old values to new
        values.
    output_col: str
        The name of the output column.
    nodata : int or float, optional
        The value used to fill the nodata records.
    Returns
    -------
    input_df : DataFrame or GeoDataFrame
        The output DataFrame with the reclassified values.

    """
    key_type = set(map(type, [*reclassify_def]))

    if key_type == {tuple}:
        # get the lowest interval
        lowest_id = np.argmin(
            [left for left, right in reclassify_def.keys()]
        ).item()
        lowest_left, lowest_right = list(reclassify_def.keys())[lowest_id]
        # change values equal to the lowest left to the lowest right
        input_df.loc[input_df[input_col] == lowest_left,
                     input_col] = lowest_right
        intervals = pd.Series(
            list(reclassify_def.values()),
            index=pd.IntervalIndex.from_tuples([*reclassify_def])
        )
        output_sr = input_df[input_col].map(intervals)
    elif key_type == {int} or key_type == {str} or key_type == {float}:
        output_sr = input_df[input_col].map(reclassify_def)
    else:
        raise ValueError("Keys of the reclassify dictionary must be "
                         "exclusively of string, number, or tuple of two "
                         "numbers.")
    if output_sr.isna().values.any() and nodata:
        output_sr.fillna(nodata, inplace=True)

    input_df[output_col] = output_sr
    return input_df


def linear(input_df, input_col, output_col,
           start=None, end=None,
           output_min=1, output_max=9):
    """
    Rescale a column in a DataFrame linearly.

    If argument start is greater than end, the rescaling is in the
    same direction as values in the input column, i.e., smaller (bigger) values
    in the input column correspond to smaller (bigger) values in the output.
    If argument start is less than end, the rescaling is in the reverse
    direction as values in the input column.
    The start and end of the input column do not necessarily to be the minimum
    and maximum of the input column. Values beyond the specified bound will be
    assigned to output_min and output_max, depending on which side they are on.

    Parameters
    ----------
    input_df : DataFrame or GeoDataFrame
        Input DataFrame containing a column need to be rescaled.
    input_col : str
        Name of the old column.
    output_col : str
        Name of the new column.
    start : int or float
        Value from which the rescaling starts.
    end : int or float
        Value from which the rescaling ends.
    output_min : int or float
        The minimum value of the output column.
    output_max : int or float
        The maximum value of the output column.

    Returns
    -------
    input_df : DataFrame or GeoDataFrame
        Output DataFrame containing the rescaled column.

    """
    if start is None:
        start = input_df[input_col].min()
    if end is None:
        end = input_df[input_col].max()
    input_range = abs(end - start)
    output_range = abs(output_max - output_min)
    if end > start:
        input_df[output_col] = (
            output_min
            + (input_df[input_col] - start) * output_range / input_range
        )
        input_df.loc[input_df[input_col] > end, output_col] = output_max
        input_df.loc[input_df[input_col] < start, output_col] = output_min
    else:
        input_df[output_col] = (
            output_max
            - (input_df[input_col] - end) * output_range / input_range
        )
        input_df.loc[input_df[input_col] < end, output_col] = output_max
        input_df.loc[input_df[input_col] > start, output_col] = output_min
    return input_df


def gamma(input_df, input_col, output_col, output_min=1, output_max=9):
    """
    Rescale values in a column based on an asymptotic gamma distribution.

    The function calls a gamma object from the stats module in the SciPy
    package, see more detail from:
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html
    First, Maximum Likelihood Estimation (MLE) is used to estimate the shape
    and scale parameters from the data. Then, the function calculates a new
    value for its associated old value, x, by scaling the probability of the
    random variable being less than or equal to x, based on the the estimated
    cumulative density function (cdf).

    Parameters
    ----------
    input_df : DataFrame or GeoDataFrame
        Input DataFrame containing a column need to be rescaled.
    input_col : str
        Name of the old column.
    output_col : str
        Name of the new column.
    output_min : int or float
        The minimum value of the output column.
    output_max : int or float
        The maximum value of the output column.

    Returns
    -------
    input_df : DataFrame or GeoDataFrame
        Output DataFrame containing the rescaled column.
    """
    df = input_df[~np.isnan(input_df[input_col])]
    a, loc, b = stats.gamma.fit(df[input_col], floc=0)
    input_df[output_col] = output_min
    new_scale = output_max - output_min
    rescale_start = output_min + 1
    input_df.loc[~np.isnan(input_df[input_col]), output_col] = (
        stats.gamma.cdf(df[input_col], a, loc=0, scale=b)
        * (new_scale - rescale_start)
        + rescale_start
    )
    return input_df



from rasterstats import zonal_stats
import pandas as pd
from affine import Affine
from geopandas import GeoDataFrame
from pylusat.base import GeoDataFrameManager
from pylusat.base import RasterManager


def _to_affine(cellsize, max_y, min_x):
    """Construct an Affine object."""
    return Affine(cellsize, 0, min_x, 0, -cellsize, max_y)


def zonal_stats_raster(zone_gdf, raster, stats=None,
                       stats_prefix='zonal', nodata=None):
    """
    Calculate specified stats for each geometry in the zone GeoDataFrame.

    If raster and zone GeoDataFrame were in different spatial reference
    systems, the raster will be warped (reprojected) to match the spatial
    reference of the zone GeoDataFrame.

    Parameters
    ----------
    zone_gdf : GeoDataFrame
        The zone GeoDataFrame whose geometry must be polygon.
    raster : str
        A path to a tif file or a connection string to a raster on PostgreSQL.
        The raster dataset whose values are summarized.
    stats : list of str, or space-delimited str, optional
        Which statistics to calculate for each zone.
        Defaults to rasterstats.utils.DEFAULT_STATS, i.e.,
        ['count', 'min', 'max', 'mean'].
        Other valid stats are ['sum', 'std', 'median', 'majority', 'minority',
        'unique', 'range', 'nodata', 'nan'].
        See rasterstats for more details.
    stats_prefix : str, default 'zonal'
        The prefix used to name the output columns of the calculated stats.
        The output column name will be concatenated by a underscore.
        e.g., 'zonal_mean'
    nodata : int or float
        Value for no data cells.

    Returns
    -------
    output_gdf : GeoDataFrame
         Returns a GeoDataFrame containing new column(s) for each type of
         statistic specified in the ``stats`` argument.
    """
    if not GeoDataFrameManager(zone_gdf).geom_type_validate("Polygon"):
        raise ValueError("zone GeoDataFrame must be polygon.")
    gdf_crs = zone_gdf.crs

    rast_manager = RasterManager(raster, nodata)
    rast_crs = rast_manager.get_rio_crs()

    if gdf_crs.to_epsg() != rast_crs.to_epsg():
        rast_arr, cellsize, max_y, min_x, nodata = (
            rast_manager.as_rebuild_info(
                projected_rast_ds=rast_manager.reproject_vrt(crs=gdf_crs)
            )
        )
    else:
        rast_arr, cellsize, max_y, min_x, nodata = (
            rast_manager.as_rebuild_info()
        )

    affine = _to_affine(cellsize, max_y, min_x)
    zonal_output = zonal_stats(vectors=zone_gdf.geometry, raster=rast_arr,
                               nodata=nodata, affine=affine, stats=stats,
                               all_touched=True)
    zone_gdf.reset_index(drop=True, inplace=True)
    if not stats:
        from rasterstats.utils import DEFAULT_STATS
        stats = DEFAULT_STATS
    elif isinstance(stats, str):
        stats = stats.split()

    col_names = [f'{stats_prefix}_{stat}' for stat in stats]
    output_gdf = zone_gdf.join(pd.DataFrame(zonal_output))
    output_gdf.rename(columns=dict(zip(stats, col_names)), inplace=True)
    return output_gdf



import geopandas as gpd
from pylusat.geotools import spatial_join, select_by_location
from pylusat.datasets import get_path
import pytest


@pytest.fixture
def acs2016_gdf():
    return gpd.read_file(get_path("acs2016"))


@pytest.fixture
def schools_gdf():
    return gpd.read_file(get_path("schools"))


def test_spatial_join(acs2016_gdf, schools_gdf):
    result = spatial_join(acs2016_gdf, schools_gdf,
                          cols_agg={'ENROLLMENT': ['sum']})
    assert result.iloc[2, -1] == 345


def test_select_by_location(acs2016_gdf, schools_gdf):
    result = select_by_location(acs2016_gdf, schools_gdf)
    assert len(result) == 75



import os

__all__ = ["available", "get_path"]
_module_path = os.path.dirname(__file__)


def _get_dataset_name_by_type(extension='.shp'):
    names = []
    for root, dirs, files in os.walk(_module_path):
        for file in files:
            if file.endswith(extension):
                names.append(os.path.basename(root))
    return names


_available_shp = _get_dataset_name_by_type('.shp')
_available_tif = _get_dataset_name_by_type('.tif')

available = _available_shp + _available_tif


def get_path(dataset):
    """
    Get the path to the data file.
    Parameters
    ----------
    dataset : str
        The name of the dataset. See ``pylusat.datasets.available`` for
        all options.
    Examples
    --------
    >>> pylusat.datasets.get_path("acs2016") # doctest: +SKIP
    '.../python3.6/site-packages/pylusat/datasets/acs2016/acs2016.shp'
    """
    if dataset in _available_shp:
        return os.path.abspath(os.path.join(_module_path,
                                            dataset,
                                            dataset + ".shp"))
    elif dataset in _available_tif:
        return os.path.abspath(os.path.join(_module_path,
                                            dataset,
                                            dataset + ".tif"))
    else:
        msg = "The dataset '{data}' is not available. ".format(data=dataset)
        msg += "Available datasets are {}".format(", ".join(available))
        raise ValueError(msg)



import geopandas as gpd
from geopandas import GeoDataFrame
import pandas as pd
from typing import Dict


def erase(input_gdf, erase_gdf=None):
    """
    Erase erase_gdf from input_gdf.

    Parameters
    ----------
    input_gdf : GeoDataFrame
        Input GeoDataFrame.
    erase_gdf : GeoDataFrame
        Erase GeoDataFrame containing the erase features.

    Returns
    -------
    output : GeoDataFrame
        The remaining features after erasure.
    """
    if erase_gdf is None:
        return input_gdf
    else:
        return gpd.overlay(
            input_gdf, gpd.GeoDataFrame({'geometry': erase_gdf.unary_union}),
            how='difference'
        )


def spatial_join(target_gdf, join_gdf, op="intersects",
                 cols_agg: Dict[str, list] = None,
                 join_type="one to one", keep_all=True):
    """
    Spatial join two GeoDataFrames.

    Parameters
    ----------
    target_gdf, join_gdf : GeoDataFrames
        The GeoDataFrame to join to the target GeoDataFrame.
    op : string, default 'intersects'
        Binary predicate, one of {'intersects', 'contains', 'within'}. See
        http://shapely.readthedocs.io/en/latest/manual.html#binary-predicates.
    cols_agg : dict, default None
        Dict of ``{column_name: a list of statistics}``, where the list of
        statistics is a list of strings containing the names of desired
        statistics for each column. Names of the statistics include:
        {'first', 'last', 'sum', 'mean', 'median', 'max', 'min',
        'std', 'var', 'count', 'size'}.
    join_type : string, default 'one to one'
        Binary predicate, one of {'one to one', 'one to many'}. The option
        'one to one' only returns one row for each target feature, whereas
        option 'one to many' return multiple rows for each match between
        target feature and join feature.
    keep_all : bool, default True
        Whether to keep all features from the target GeoDataFrame.
    Returns
    -------
    GeoDataFrame
        A GeoDataFrame contains all columns in the target GeoDataFrame and the
        specified columns from the join GeoDataFrame.
    """
    how = 'left' if keep_all else 'inner'
    gpd_sjoin = gpd.sjoin(target_gdf, join_gdf, how=how, op=op)

    if join_type.lower() == "one to one":
        sjoin_by_index = gpd_sjoin.groupby(gpd_sjoin.index)

        if cols_agg is None:
            cols_agg = {col: ['first'] for col in join_gdf.columns
                        if col != join_gdf.geometry.name}
            join_df = sjoin_by_index.agg(cols_agg)
            join_df.columns = cols_agg.keys()
        else:
            join_df = sjoin_by_index.agg(cols_agg)
            join_df.columns = [f"{key}_{v}"
                               for key, value in cols_agg.items()
                               for v in set(value)]

        # grab columns of target_gdf
        target_gdf_columns = gpd_sjoin.columns[:len(target_gdf.columns)]
        # remove duplicated rows generated by geopandas spatial join
        # based on columns of target_gdf
        target_df = gpd_sjoin[target_gdf_columns].drop_duplicates()
        return pd.concat([target_df, join_df], axis=1)
    elif join_type.lower() == "one to many":
        return gpd_sjoin
    else:
        raise ValueError("join_type must be either 'one to one' or "
                         "'one to many'")


def within_dist(input_gdf, input_id, distance,
                target_gdf=None, output_col=None):
    """
    For each object in the input, test if any object in the target set is
    within a specified distance.

    Parameters
    ----------
    input_gdf : GeoDataFrame
        The input set.
    input_id : str
        The name of the column containing the unique id of the input set.
    distance : int or float
        Distance (in the same unit as the input GeoDataFrame).
    target_gdf : GeoDataFrame
        The target set.
    output_col : str
        The name of the output column.

    Returns
    -------
    output : GeoDataFrame
        The output value is 1 if there exists any target object within the
        specified distance of the input object and 0 otherwise.
    """
    if output_col is None:
        output_col = "within_" + str(distance)
    input_gdf[output_col] = 0
    if len(target_gdf) > 0:
        sjoin_result = gpd.sjoin(
            input_gdf, target_gdf.assign(geom=target_gdf.buffer(distance)),
            how='inner', op='intersects'
        )[input_id]
        input_gdf.loc[input_gdf[input_id].isin(sjoin_result), output_col] = 1
    return input_gdf


def select_by_location(input_gdf, select_gdf,
                       op='intersects', within_dist=0):
    """
    Select part of the input GeoDataFrame based on its relationship with the
    selecting GeoDataFrame.

    Parameters
    ----------
    input_gdf : GeoDataFrame
        The input GeoDataFrame.
    select_gdf : GeoDataFrame
        The selecting GeoDataFrame.
    op : string, default 'intersection'
        Binary predicate, one of {'intersects', 'contains', 'within',
        'within a distance'}. See
        http://shapely.readthedocs.io/en/latest/manual.html#binary-predicates.
    within_dist : int, default 0
        Search distance around the select_gdf. This parameter is only
        useful when op is set to be "within a distance".
    Returns
    -------
    output : GeoDataFrame
        The selected features from the input GeoDataFrame.
    """
    ops = ['intersects', 'contains', 'within', 'within a distance']
    assert op in ops, 'invalid op parameter,'
    if op == 'within a distance' and within_dist:
        select_gdf[select_gdf.geometry.name] = select_gdf.buffer(within_dist)
        op = 'within'
    output_gdf = input_gdf.loc[
                 input_gdf.index.to_series().isin(
                     gpd.sjoin(
                         input_gdf, select_gdf,
                         how='inner', op=op
                     ).index.values
                 ), :
                 ]
    output_gdf = output_gdf.rename_axis(None, axis=1)
    return output_gdf.copy()



import numpy as np
from scipy.spatial import cKDTree
from pandas import Series
from pylusat.utils import cntrd_array


def idw(input_gdf, value_gdf, value_clm, power=2, n_neighbor=12,
        search_radius=None, leafsize=14, min_dist=1e-12, dtype=float):
    """
    Interpolation using inverse distance weighting (IDW).

    This function implements an `IDW interpolation
    <https://en.wikipedia.org/wiki/Inverse_distance_weighting>`. The power
    parameter dictates how fast the influence to a given location by its
    nearby objects decays. `idw_cv`, a k-fold cross validation method is
    offered to determine the most appropriate value of the `power` parameter.

    Parameters
    ----------
    input_gdf : GeoDataFrame
        Input GeoDataFrame. Centroids of the input geometries are used.
    value_gdf : GeoDataFrame
        GeoDataFrame containing the values needed to be interpolated.
    value_clm : str
        The name of the column that holds the values in value_gdf.
    power : int or float, optional
        The power parameter in IDW.
    n_neighbor : int, optional
        Number of neighborhoods used for IDW.
    search_radius : float, optional
        Maximum distance used to find neighbors. If not provided, the function
        will search for all neighbors specified by n_neighbors.
    leafsize : positive int, optional
        The number of points at which the algorithm switches over to
        brute-force. Default: 14. See `scipy.spatial.cKDTree` for further
        information.
    min_dist : float, optional
        The distance below which the interpolated value will be set to equal to
        the value of its closest neighbor.
    dtype : str or np.dtype, optional
        Use a np.dtype or Python type to cast the interpolated values to the
        desired type.

    Returns
    -------
    output_sr : Series
        pandas Series that contains the interpolated values for all feature in
        the input_gdf.
    """
    if not (isinstance(n_neighbor, int) and
            1 <= n_neighbor <= len(value_gdf.index)):
        # number of neighbors <= number of points created the tree
        raise ValueError("n_neighbor must be a positive integer that is less "
                         "than or equal to the number of rows in value_gdf.")

    value_gdf = value_gdf.reset_index(drop=True)
    value_coords = cntrd_array(value_gdf)
    kdtree = cKDTree(value_coords, leafsize=leafsize)

    if not search_radius:
        search_radius = np.inf
    dd: np.ndarray
    ii: np.ndarray
    dd, ii = kdtree.query(cntrd_array(input_gdf), k=n_neighbor,
                          distance_upper_bound=search_radius)

    if n_neighbor == 1:
        return value_gdf[value_clm][ii]
    if dtype is None:
        dtype = value_gdf[value_clm].dtype

    n = len(input_gdf.index)
    output_arr = np.zeros(n, dtype=dtype)
    if min_dist <= 0:
        min_dist = 1e-12

    for j in range(n):
        within_min = (dd[j] <= min_dist)
        if np.any(within_min):
            output_arr[j] = value_gdf[value_clm][ii[j][within_min][0]]
            continue

        w = 1 / dd[j]**power
        output_arr[j] = np.dot(w, value_gdf[value_clm][ii[j]]) / np.sum(w)

    output_sr = Series(output_arr, index=input_gdf.index)
    return output_sr


def idw_cv(data, yname, func, k=10, seed=None, **kwargs):
    n = len(data.index)  # total number of observations
    f = int(np.ceil(n/k))  # sample size in each fold
    np.random.seed(seed)
    # randomly assign observations to the k-th fold
    s = np.random.choice(np.tile(np.arange(k), f), n, replace=False)

    data = data.reset_index(drop=True)
    mse = np.zeros(k)   # mean square error
    for i in range(k):
        data_test = data.loc[s == i, ]
        data_train = data.loc[s != i, ]
        observed_y = data_test[yname]
        fitted_y = func(data_test, data_train, yname, **kwargs)
        mse[i] = np.mean(np.power((observed_y - fitted_y), 2))
    return np.mean(mse)



import geopandas as gpd
from pylusat.interpolate import idw
from pylusat.datasets import get_path
import pytest


@pytest.fixture
def acs2016_gdf():
    return gpd.read_file(get_path("acs2016"))


@pytest.fixture
def schools_gdf():
    return gpd.read_file(get_path("schools"))


def test_idw(acs2016_gdf, schools_gdf):
    idw_result = idw(acs2016_gdf, schools_gdf, 'ENROLLMENT',
                     power=2.00, n_neighbor=12)
    assert round(idw_result[0], 4) == 26.4073



from pyproj import Proj
from geopandas import GeoDataFrame
import numpy as np
import rasterio as rio
from rasterio.vrt import WarpedVRT


class GeoDataFrameManager:

    def __init__(self, gdf_obj):
        assert type(gdf_obj) == GeoDataFrame, ("Your input is not a valid "
                                               "GeoDataFrame.")
        self.gdf = gdf_obj

    def geom_type_validate(self, validate_type):
        type_sr = self.gdf.geom_type
        validation_map = {
            "Point": type_sr.isin(["Point", "MultiPoint"]).all(),
            "Line": type_sr.isin(["LineString", "MultiLineString"]).all(),
            "Polygon": type_sr.isin(["Polygon", "MultiPolygon"]).all()
        }
        return validation_map[validate_type]

    @property
    def geom_unit_id(self) -> str:
        proj_str = Proj(self.gdf.crs).definition_string()
        return [_[_.find("=") + 1:] for _ in proj_str.split()
                if _.startswith("units=")][0]

    @property
    def geom_unit_name(self):
        return UnitHandler(self.geom_unit_id).fullname

    @classmethod
    def from_shp(cls, shp_path):
        assert shp_path.endswith(".shp"), "Not a valid shapefile."
        try:
            return cls(GeoDataFrame.from_file(shp_path))
        except Exception as err:
            print(err)


class UnitHandler:

    UNIT_MAP = {'km': ('Kilometer', 1000.0),
                'm': ('Meter', 1.0),
                'dm': ('Decimeter', 0.1),
                'cm': ('Centimeter', 0.01),
                'mm': ('Millimeter', 0.001),
                'kmi': ('International Nautical Mile', 1852.0),
                'in': ('International Inch', 0.0254),
                'ft': ('International Foot', 0.3048),
                'yd': ('International Yard', 0.9144),
                'mi': ('International Statute Mile', 1609.344),
                'fath': ('International Fathom', 1.8288),
                'ch': ('International Chain', 20.1168),
                'link': ('International Link', 0.201168),
                'us-in': ("U.S. Surveyor's Inch", 0.0254000508001016),
                'us-ft': ("U.S. Surveyor's Foot", 0.3048006096012192),
                'us-yd': ("U.S. Surveyor's Yard", 0.9144018288036576),
                'us-ch': ("U.S. Surveyor's Chain", 20.116840233680467),
                'us-mi': ("U.S. Surveyor's Statute Mile", 1609.3472186944373),
                'ind-yd': ('Indian Yard', 0.91439523),
                'ind-ft': ('Indian Foot', 0.30479841),
                'ind-ch': ('Indian Chain', 20.11669506)}

    # alternative unit names
    OTHER_NAMES = {"Metre": "m",
                   "Mile": "us-mi",
                   "Foot": "us-ft",
                   "Yard": "us-yd",
                   "Inch": "us-in"}

    # normal unit names
    VALID_NAMES = {v[0]: k for k, v in UNIT_MAP.items()}

    # additional area names supported
    AREA_NAMES = {'Acre': 'ac',
                  'Hectare': 'ha'}

    def __init__(self, unit):
        validate_1d = self._validate_1d(unit)
        if validate_1d:
            self.unit_id = validate_1d
            self.dimension = 1
            return
        validate_2d = self._validate_2d(unit)
        if validate_2d:
            self.unit_id = validate_2d
            self.dimension = 2
            return
        else:
            raise ValueError(f'{unit} is not a valid unit.')

    @property
    def fullname(self):
        if self.dimension == 1:
            return self.UNIT_MAP[self.unit_id][0]
        else:  # dimension=2
            if self.unit_id == 'ac':
                return 'Acre'
            elif self.unit_id == 'ha':
                return 'Hectare'
            else:
                return f'Square {self.UNIT_MAP[self.unit_id][0]}'

    @staticmethod
    def _pluralize(name):
        # convert unit name to its corresponding plural form
        if name.endswith("Inch"):
            return name.replace("Inch", "Inches")
        elif name.endswith("Foot"):
            return name.replace("Foot", "Feet")
        else:
            return name + "s"

    @property
    def plural(self):
        return self._pluralize(self.fullname)

    @property
    def base_factor(self):
        """conversion factor compare to meter or square meter."""
        if self.dimension == 1:
            return self.UNIT_MAP[self.unit_id][1]
        else:  # dimension=2
            if self.unit_id == 'ac':
                return self.UNIT_MAP['us-ft'][1]**2 * 43560
            elif self.unit_id == 'ha':
                return 10000
            else:
                return self.UNIT_MAP[self.unit_id][1]**2

    def _unit_name_map(self):
        return {**self.VALID_NAMES, **self.OTHER_NAMES,
                **{self._pluralize(k): v for k, v in self.VALID_NAMES.items()},
                **{self._pluralize(k): v for k, v in self.OTHER_NAMES.items()}}

    def _validate_1d(self, unit):
        try:
            unit_lower = unit.lower()
            unit_title = unit.title()
            unit_name_map = self._unit_name_map()
            if unit_lower in self.UNIT_MAP.keys():
                return unit_lower
            elif unit_title in unit_name_map.keys():
                return unit_name_map[unit_title]
            else:
                return None
        except Exception as err:
            print(err)

    def _validate_2d(self, unit):
        try:
            sq, unit_name = unit.split()
            if sq.lower() in ['square', 'sq'] and self._validate_1d(unit_name):
                return self._validate_1d(unit_name)
            else:
                raise ValueError('Not a valid area unit.')
        except ValueError:
            if unit.title() in self.AREA_NAMES.keys():
                return self.AREA_NAMES[unit.title()]
            elif unit[:-1].title() in self.AREA_NAMES.keys():
                return self.AREA_NAMES[unit[:-1].title()]
            else:
                raise ValueError('Not a valid area unit.')

    def convert(self, other_unit):
        other_unit_handler = UnitHandler(other_unit)
        if self.dimension != other_unit_handler.dimension:
            raise TypeError(f'Incompatible units, from {self.fullname} to '
                            f'{other_unit_handler.fullname}.')
        else:
            return self.base_factor / other_unit_handler.base_factor


class RasterManager:

    def __init__(self, rast_file, nodata=None):
        self.rast_file = rast_file
        self.rast_ds = self._validate_rast()
        self.rast_nodata = nodata

    def _validate_rast(self):
        try:
            rast_ds = rio.open(self.rast_file)
            return rast_ds
        except Exception:
            raise ValueError("Not a valid raster data.")

    def get_rio_dataset(self):
        # get rasterio dataset
        return self.rast_ds

    def get_rio_crs(self):
        # get rasterio.crs from the raster dataset
        return self.rast_ds.crs

    def to_array(self):
        # return a numpy.ndarray
        return self.rast_ds.read(1)

    def get_affine(self):
        return self.rast_ds.transform

    def as_rebuild_info(self, projected_rast_ds=None):
        if not projected_rast_ds:
            rast_ds = self.rast_ds
        else:
            rast_ds = projected_rast_ds
        rast_band = rast_ds.read(1)
        rast_affine = rast_ds.transform
        if self.rast_nodata is not None:
            rast_band[
                np.where(rast_band == rast_ds.nodata)
            ] = self.rast_nodata
            nodata = self.rast_nodata
        else:
            nodata = rast_ds.nodata
        cellsize = rast_affine[0]
        min_x = rast_affine[2]
        max_y = rast_affine[5]
        return rast_band, cellsize, max_y, min_x, nodata

    @property
    def wkt(self):
        return self.rast_ds.crs.to_wkt()

    def reproject_vrt(self, crs=None):
        return WarpedVRT(self.rast_ds, crs)



import geopandas as gpd
from pylusat.utils import gridify
from pylusat.datasets import get_path
import pytest


@pytest.fixture
def schools_gdf():
    return gpd.read_file(get_path("schools"))


def test_gridify(schools_gdf):
    result = gridify(schools_gdf, width=1000)
    assert len(result) == 2496



"""Setup for the pylusat package."""

import setuptools
import os
from pylusat._version import __version__

with open('README.md') as f:
    README = f.read()

data_files = []

for item in os.listdir("pylusat/datasets"):
    if not item.startswith("__"):
        data_files.append(os.path.join("datasets", item, "*"))

setuptools.setup(
    name='pylusat',
    author="Changjie Chen",
    author_email="chj.chen@hotmail.com",
    license="BSD",
    description='Python for Land-use suitability analysis tools',
    version=__version__,
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/chjch/pylusat',
    packages=setuptools.find_packages(exclude=("tests",)),
    python_requires=">=3.6",
    install_requires=["geopandas", "rasterio", "rasterstats", "scipy"],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    package_data={"pylusat": data_files}
)



import geopandas as gpd
from pylusat.rescale import linear
from pylusat.datasets import get_path
import pytest


@pytest.fixture
def schools_gdf():
    return gpd.read_file(get_path("schools"))


def test_rescale(schools_gdf):
    result = linear(schools_gdf, "ENROLLMENT", "ENROLL_CLS")
    assert round(result.iloc[0, -1], 4) == 1.3849



import geopandas as gpd
from pylusat.density import of_point, of_line
from pylusat.datasets import get_path
import pytest


@pytest.fixture
def acs2016_gdf():
    return gpd.read_file(get_path("acs2016"))


@pytest.fixture
def schools_gdf():
    return gpd.read_file(get_path("schools"))


@pytest.fixture
def highway_gdf():
    return gpd.read_file(get_path("highway"))


def test_of_point_no_search(acs2016_gdf, schools_gdf):
    result = of_point(acs2016_gdf, schools_gdf, area_unit='square mile')
    assert round(result[0], 4) == 2.3894


def test_of_point_search(acs2016_gdf, schools_gdf):
    result = of_point(acs2016_gdf, schools_gdf,
                      "ENROLLMENT", '1 mile', 'square mile')
    assert round(result[1], 4) == 0.0003


def test_of_line(acs2016_gdf, highway_gdf):
    result = of_line(acs2016_gdf, highway_gdf,
                     search_radius='1 mile',
                     area_unit='square mile')
    assert round(result[0], 4) == 0.0004



import pandas as pd
import numpy as np
import geopandas as gpd
from geopandas import GeoDataFrame
from pylusat.base import GeoDataFrameManager, UnitHandler


def _buffer(input_gdf, buffer_dist):
    """create buffers around the input_gdf using specified buffer distance"""

    input_unit = GeoDataFrameManager(input_gdf).geom_unit_id
    buff_length, buff_unit = buffer_dist.split()
    # calculate the length of buffer in the unit used by input_gdf
    buff_factor = float(buff_length)/UnitHandler(input_unit).convert(buff_unit)
    return input_gdf.buffer(buff_factor)


def _buffer_factor(input_gdf, buffer_dist):
    """create buffers around the input_gdf using specified buffer distance"""

    input_unit = GeoDataFrameManager(input_gdf).geom_unit_id
    buff_length, buff_unit = buffer_dist.split()
    # calculate the length of buffer in the unit used by input_gdf
    buff_factor = float(buff_length)/UnitHandler(input_unit).convert(buff_unit)
    return buff_factor


def of_point(input_gdf, point_gdf, pop_clm=None,
             search_radius=None, area_unit='square meters'):
    """
    Calculate density of points in each input geometry.

    Parameters
    ----------
    input_gdf : GeoDataFrame
        Input GeoDataFrame in which points are counted.
    point_gdf : GeoDataFrame
        Point GeoDataFrame.
    pop_clm : str
        Population column which contains values to represent each occurrence
        of the point feature.
    search_radius : str, optional
        A string of buffering distance and unit, separated by space.
        e.g., "1 mile".
    area_unit : str, optional
        A string of the area unit used for density calculation.
        e.g., "square meters".

    Returns
    -------
    output_sr : pd.Series
        A pandas Series that contains the density of point in each input
        geometry of the input GeoDataFrame.
    """
    input_gdf_manager = GeoDataFrameManager(input_gdf)
    point_gdf_manager = GeoDataFrameManager(point_gdf)
    assert point_gdf_manager.geom_type_validate("Point"), (
        "The geometry of point GeoDataFrame must be Point."
    )

    if not search_radius:
        assert input_gdf_manager.geom_type_validate("Polygon"), (
            "The geometry of input GeoDataFrame must be Polygon "
            "if `search_radius` is None."
        )
        search_unit = input_gdf_manager.geom_unit_id
        input_copy = input_gdf
    else:
        input_copy = input_gdf.copy()
        input_copy.geometry = _buffer(input_copy, search_radius)
        search_unit = search_radius.split()[1]

    joint_gdf = gpd.sjoin(input_copy, point_gdf, how='left', op='intersects')
    by_input_index = joint_gdf.groupby(level=0)
    if pop_clm is None:
        output_sr = by_input_index["index_right"].count()
    else:
        output_sr = by_input_index[pop_clm].sum()
    output_sr.fillna(0)
    output_sr.index = input_gdf.index

    search_area = input_copy.area
    # convert search area to specified areal unit
    search_area *= UnitHandler(f'square {search_unit}').convert(area_unit)

    output_sr = output_sr / search_area
    return output_sr


def of_line(input_gdf, line_gdf, cellsize=30, search_radius=None,
            area_unit="square meters", geomtoarray=None):
    """
    Calculate density of line length in each input geometry.

    Parameters
    ----------
    input_gdf : GeoDataFrame
        Input GeoDataFrame in which length of lines are summed.
    line_gdf : GeoDataFrame
        Line GeoDataFrame whose lengths are summed.
    cellsize : float, optional
        The cell size used to rasterize the line_gdf.
    search_radius : str, optional
        A string of buffering distance and unit, separated by space.
        e.g., "1 mile".
    area_unit : str, optional
        A string of the area unit used for density calculation.
        e.g., "square meters".
    geomtoarray : {tuple of (np.ndarray, Affine, np.ndarray, float), None}
        The output of ``rasterize_geometry`` function. If a desired output has
        already been created for the `line_gdf` has already been created,
        set it to be this argument to use it.

    Returns
    -------
    output_sr : pd.Series
        A pandas Series that contains the density of line in each input
        geometry of the input GeoDataFrame.
    """
    from pylusat.utils import rasterize_geometry
    from rasterstats import zonal_stats
    import warnings

    input_gdf_manager = GeoDataFrameManager(input_gdf)
    line_gdf_manager = GeoDataFrameManager(line_gdf)
    assert line_gdf_manager.geom_type_validate("Line"), (
        "The geometry of line GeoDataFrame must be Line."
    )

    line_data = rasterize_geometry(line_gdf, cellsize) if geomtoarray is None \
        else geomtoarray

    if not search_radius:
        assert input_gdf_manager.geom_type_validate("Polygon"), (
            "The geometry of input GeoDataFrame must be Polygon, "
            "if `search_radius` is None."
        )
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            zstats = zonal_stats(input_gdf, line_data[0], affine=line_data[1],
                                 stats=['sum'], nodata=0)
        search_area = input_gdf.area
    else:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            buff_factor = _buffer_factor(input_gdf, search_radius)
            input_buff = input_gdf.centroid.buffer(buff_factor)
            zstats = zonal_stats(input_buff, line_data[0], affine=line_data[1],
                                 stats=['sum'], nodata=0)
        search_unit = search_radius.split()[1]
        search_area = input_buff.area
        # convert search area to specified areal unit
        search_area *= UnitHandler(f'square {search_unit}').convert(area_unit)

    output_arr = np.array([v if v is not None else np.nan
                           for d in zstats
                           for v in d.values()])
    return pd.Series(output_arr*cellsize/search_area, index=input_gdf.index)
